
#include <stdint.h>
#include <unistd.h>
#include <time.h>
#include <sys/times.h>

#include <stdio.h>

#include <result.h>
#include <serial_io.h>

#if defined(__CYGWIN__)
	// Compiling under windows
	// Required libraries: 
	#include <windows.h>
	#include <string.h>
	// POSIX serial IO seems to work very slowly and unpredictably. Trying native Windows instead.
	
#elif defined(__DARWIN__)
	// Compiling under Mac OSX
	// Required libraries:
	#include <fcntl.h>   /* File control definitions */
	#include <termios.h> /* POSIX terminal control definitions */
	
#elif defined(__linux__)
	// Compiling under Linux
	// Required libraries: 
	#include <fcntl.h>   /* File control definitions */
	#include <termios.h> /* POSIX terminal control definitions */
	
#endif





//==================================================================================================
// Cygwin Version (Windows)
//==================================================================================================
#if defined(__CYGWIN__)

#define SERIAL_TIMEOUT_TICKS	((SERIAL_TIMEOUT*CLOCKS_PER_SEC)/1000)
	HANDLE hSerial;
	
RES_t serialOpen(char *strdev){
	COMMTIMEOUTS timeouts={0};
	char strdev2[16];

	sprintf(strdev2,"\\\\.\\%s",strdev);
	hSerial = CreateFile(strdev2,
						GENERIC_READ | GENERIC_WRITE,
						0,
						0,
						OPEN_EXISTING,
						FILE_ATTRIBUTE_NORMAL,
						0);
						
	if(hSerial==INVALID_HANDLE_VALUE){
		return(RES_FAIL);
	}
	
	// Turn off timeouts. R/W functions are nonblocking.
	timeouts.ReadIntervalTimeout=MAXDWORD;
	timeouts.ReadTotalTimeoutConstant=0;
	timeouts.ReadTotalTimeoutMultiplier=0;
	timeouts.WriteTotalTimeoutConstant=0;
	timeouts.WriteTotalTimeoutMultiplier=0;
	if(!SetCommTimeouts(hSerial, &timeouts)){
		return(RES_FAIL);
	}
	
	return(RES_OK);
}

//--------------------------------------------------------------------------------------------------
RES_t serialWrite(void *buf, size_t size){
	clock_t clocks1, clocks2;
	struct tms timebuf;
	DWORD nbytes;
	
	clocks1 = times(&timebuf);
	while(size){
		WriteFile(hSerial,buf,size,&nbytes,NULL);
		if(nbytes > 0){
			buf += nbytes;
			size -= nbytes;
			clocks1 = times(&timebuf);
		}
		
		// see if timed out
		clocks2 = times(&timebuf);
		if(clocks2 > (clocks1+SERIAL_TIMEOUT_TICKS)) return(RES_FAIL);
	}
	
	return(RES_OK);
}

//--------------------------------------------------------------------------------------------------
RES_t serialRead(void *buf, size_t size){
	clock_t clocks1, clocks2;
	struct tms timebuf;
	DWORD nbytes;
	
	clocks1 = times(&timebuf);
	while(size){
		ReadFile(hSerial,buf,size,&nbytes,NULL);
		if(nbytes > 0){
			buf += nbytes;
			size -= nbytes;
			clocks1 = times(&timebuf);
		}
		
		// see if timed out
		clocks2 = times(&timebuf);
		if(clocks2 > (clocks1+SERIAL_TIMEOUT_TICKS)) return(RES_FAIL);
	}
	
	return(RES_OK);
}

//--------------------------------------------------------------------------------------------------
void serialFlushInput(void){
	PurgeComm(hSerial,PURGE_RXCLEAR);
}
//--------------------------------------------------------------------------------------------------
RES_t serialClose(void){
	CloseHandle(hSerial);
	return(RES_OK);
}

//==================================================================================================
// Linux/Mac OSX Version
//==================================================================================================
#elif defined(__linux__) || defined(__DARWIN__)

#define SERIAL_TIMEOUT_TICKS	((SERIAL_TIMEOUT*CLOCKS_PER_SEC)/1000)
static int tty_dev;

RES_t serialOpen(char *strdev){
	struct termios options;
	
	tty_dev = open(strdev, O_RDWR | O_NOCTTY | O_NDELAY);
	if(tty_dev == -1){
		return(RES_FAIL);
	}
	
	fcntl(tty_dev, F_SETFL, FNDELAY); // enable nonblocking reads
	
	// Set IO to no parity, no flow control, no pre/post processing.
	tcgetattr(tty_dev, &options);
	options.c_iflag &= ~(BRKINT|ICRNL|IGNBRK|IGNCR|IGNPAR|INLCR|INPCK|ISTRIP|IXANY|IXOFF|PARMRK);
	options.c_oflag &= ~(OPOST|ONLCR|OCRNL|ONOCR|ONLRET|OFILL|NLDLY|CRDLY|TABDLY|BSDLY|VTDLY|FFDLY);
	options.c_lflag &= ~(ECHO|ECHOE|ECHOK|ECHONL|ICANON|IEXTEN|ISIG|NOFLSH|TOSTOP);
	options.c_cflag &= ~(CSIZE|CSTOPB|PARENB|CRTSCTS);
	options.c_cflag |= CS8|CLOCAL;
	tcsetattr(tty_dev, TCSANOW, &options);
	
	
	return(RES_OK);
}

//--------------------------------------------------------------------------------------------------
RES_t serialWrite(void *buf, size_t size){
	clock_t clocks1, clocks2;
	struct tms timebuf;
	ssize_t wrcount;
	
	clocks1 = times(&timebuf);
	while(size){
		wrcount = write(tty_dev,buf,size);
		if(wrcount > 0){
			buf += wrcount;
			size -= wrcount;
			clocks1 = times(&timebuf);
		}
		
		// see if timed out
		clocks2 = times(&timebuf);
		if(clocks2 > (clocks1+SERIAL_TIMEOUT_TICKS)) return(RES_FAIL);
	}
	
	return(RES_OK);
}

//--------------------------------------------------------------------------------------------------
RES_t serialRead(void *buf, size_t size){
	clock_t clocks1, clocks2;
	struct tms timebuf;
	ssize_t rdcount;
	
	clocks1 = times(&timebuf);
	while(size){
		rdcount = read(tty_dev,buf,size);
		if(rdcount > 0){
			buf += rdcount;
			size -= rdcount;
			clocks1 = times(&timebuf);
		}
		
		// see if timed out
		clocks2 = times(&timebuf);
		if(clocks2 > (clocks1+SERIAL_TIMEOUT_TICKS)) return(RES_FAIL);
	}
	
	return(RES_OK);
}

//--------------------------------------------------------------------------------------------------
void serialFlushInput(void){
	tcflush(tty_dev, TCIFLUSH);
}

//--------------------------------------------------------------------------------------------------
RES_t serialClose(void){
	close(tty_dev);
	return(RES_OK);
}

#endif
